"""
Test and validate the creation of variational parameters and circuit parameters. 
"""
import unittest
import numpy as np
from scipy.fft import dst, dct

from openqaoa.qaoa_components import *
from openqaoa.utilities import X_mixer_hamiltonian, XY_mixer_hamiltonian

from openqaoa.qaoa_components.variational_parameters.variational_params_factory import PARAMS_CLASSES_MAPPER

register = [0, 1, 2]
terms = [[0, 1], [2], [0, 2]]
weights = [1, 0.5, -2.0]

terms_wo_bias = [[0, 1], [1, 2]]
weights_wo_bias = [1, 0.5]

cost_hamiltonian = Hamiltonian.classical_hamiltonian(
    terms, weights, constant=0)
cost_hamiltonian_wo_bias = Hamiltonian.classical_hamiltonian(terms_wo_bias,
                                                             weights_wo_bias,
                                                             constant=0.7)
mixer_hamiltonian = X_mixer_hamiltonian(len(register))


class TestingQAOADescriptor(unittest.TestCase):
    
    def setUp(self):
        
        self.p = 2
        self.cost_hamil = Hamiltonian([PauliOp('ZZ', (0, 1)), PauliOp('ZZ', (1, 2)), 
                                       PauliOp('ZZ', (0, 2)), PauliOp('Z', (0, ))], 
                                      [1, 1, 1, 0.5], 1)
        self.mixer_hamil = X_mixer_hamiltonian(n_qubits = 3)
        self.mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2)]
        self.mixer_gatemap_coeffs = self.mixer_hamil.coeffs
    
    def test_QAOADescriptor(self):
        
        """
        QAOADescriptor accept 2 types of inputs for mixer_block argument on initilisation. This test checks that the same mixer block when presented in Hamiltonian or as a List of RotationGateMap, plus the appropriate mixer coefficients, produces similar internal attributes.
        """
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p=self.p)
        
        qaoa_descriptor_gm = QAOADescriptor(self.cost_hamil, self.mixer_gatemap, p=self.p, 
                                              mixer_coeffs = self.mixer_gatemap_coeffs)
        
        # Check same number of mixer gates
        self.assertEqual(len(qaoa_descriptor.mixer_block), self.p)
        self.assertEqual(len(qaoa_descriptor_gm.mixer_block), self.p)
        for each_p_value in range(self.p):
            self.assertEqual(len(qaoa_descriptor.mixer_block[each_p_value]), 3)
            self.assertEqual(len(qaoa_descriptor_gm.mixer_block[each_p_value]), 3)
        # Check similar coefficients and qubit singles/pairs names
        self.assertEqual(qaoa_descriptor.mixer_single_qubit_coeffs, 
                         qaoa_descriptor_gm.mixer_single_qubit_coeffs)
        self.assertEqual(qaoa_descriptor.mixer_pair_qubit_coeffs, 
                         qaoa_descriptor_gm.mixer_pair_qubit_coeffs)
        self.assertEqual(qaoa_descriptor.mixer_qubits_singles, 
                         qaoa_descriptor_gm.mixer_qubits_singles)
        self.assertEqual(qaoa_descriptor.mixer_qubits_pairs, 
                         qaoa_descriptor_gm.mixer_qubits_pairs)
        
    def test_QAOADescriptor_mixer_coeffs_selector_hamiltonian(self):
        
        """
        If a mixer hamiltonian is used for mixer_block, the coefficients will be obtained from it. Even if the user inputs his own mixer coefficient, it will be ignored.
        """
        
        verify_mixer_coeffs = [-1, -1, -1]
        
        mixer_hamil = X_mixer_hamiltonian(n_qubits = 3)
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p = self.p)
        qaoa_descriptor_2 = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p = self.p, 
                                             mixer_coeffs = [1, 0, 1])
        
        self.assertEqual(qaoa_descriptor.mixer_block_coeffs, verify_mixer_coeffs)
        self.assertEqual(qaoa_descriptor_2.mixer_block_coeffs, verify_mixer_coeffs)
    
    def test_QAOADescriptor_mixer_coeffs_selector_gatemap(self):
        
        """
        If a mixer gatemap is used for mixer_block, the user is required to input his own mixer coefficients, there should be an error message if the user does not input the appropriate number of mixer coefficients.
        """
        
        verify_mixer_coeffs = [-1, -1, -1]
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_gatemap, p = self.p, 
                                           mixer_coeffs = self.mixer_gatemap_coeffs)
        
        self.assertEqual(qaoa_descriptor.mixer_block_coeffs, verify_mixer_coeffs)
        
        with self.assertRaises(ValueError) as cm:
            QAOADescriptor(self.cost_hamil, self.mixer_gatemap, self.p, [])
            self.assertEqual('The number of terms/gatemaps must match the number of coefficients provided.', str(cm.exception))
    
    def test_QAOADescriptor_assign_coefficients(self):
        
        """
        The method should split the coefficients and gatemaps in the proper order. Regardless of their positions within the hamiltonian or gatemap list.
        """
        
        verify_mixer_coeffs = [-1, -1, -1, -0.5]
        
        
        # TestCase 1: GateMap Mixer
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, mixer_gatemap, p = self.p, 
                                           mixer_coeffs = [-1, -1, -1, -0.5])
        
        self.assertEqual(qaoa_descriptor.cost_single_qubit_coeffs, [0.5])
        self.assertEqual(qaoa_descriptor.cost_pair_qubit_coeffs, [1, 1, 1])
        self.assertEqual(qaoa_descriptor.cost_qubits_singles, ['RZGateMap'])
        self.assertEqual(qaoa_descriptor.cost_qubits_pairs, ['RZZGateMap', 'RZZGateMap', 'RZZGateMap'])
        
        self.assertEqual(qaoa_descriptor.mixer_single_qubit_coeffs, [-1, -1, -1])
        self.assertEqual(qaoa_descriptor.mixer_pair_qubit_coeffs, [-0.5])
        self.assertEqual(qaoa_descriptor.mixer_qubits_singles, ['RXGateMap', 'RXGateMap','RXGateMap'])
        self.assertEqual(qaoa_descriptor.mixer_qubits_pairs, ['RXXGateMap'])
        
        # TestCase 2: Hamiltonian Mixer
        mixer_gatemap = XY_mixer_hamiltonian(n_qubits = 3)
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, mixer_gatemap, p = self.p)
        
        self.assertEqual(qaoa_descriptor.cost_single_qubit_coeffs, [0.5])
        self.assertEqual(qaoa_descriptor.cost_pair_qubit_coeffs, [1, 1, 1])
        self.assertEqual(qaoa_descriptor.cost_qubits_singles, ['RZGateMap'])
        self.assertEqual(qaoa_descriptor.cost_qubits_pairs, ['RZZGateMap', 'RZZGateMap', 'RZZGateMap'])
        
        self.assertEqual(qaoa_descriptor.mixer_single_qubit_coeffs, [])
        self.assertEqual(qaoa_descriptor.mixer_pair_qubit_coeffs, 
                         [0.5, 0.5, 0.5, 0.5 , 0.5, 0.5])
        self.assertEqual(qaoa_descriptor.mixer_qubits_singles, [])
        self.assertEqual(qaoa_descriptor.mixer_qubits_pairs, 
                         ['RXXGateMap', 'RYYGateMap', 'RXXGateMap', 'RYYGateMap', 'RXXGateMap', 'RYYGateMap'])
    
    def test_QAOADescriptor_cost_block(self):
        
        """
        cost_block property should always return a list of RotationGateMaps based on the input cost hamiltonian.
        """
        
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, mixer_gatemap, p = self.p, 
                                           mixer_coeffs = [-1, -1, -1, -0.5])
        
        for each_p_block in qaoa_descriptor.cost_block:
            for each_item in each_p_block:
                self.assertTrue(isinstance(each_item, RotationGateMap))
    
    def test_QAOADescriptor_mixer_block(self):
        
        """
        mixer_block property should always return a list of RotationGateMaps based on the input cost hamiltonian.
        """
        
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2), RXXGateMap(0, 2)]
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, mixer_gatemap, p = self.p, 
                                           mixer_coeffs = [-1, -1, -1, -0.5])
        
        for each_p_block in qaoa_descriptor.mixer_block:
            for each_item in each_p_block:
                self.assertTrue(isinstance(each_item, RotationGateMap))
                
    def test_QAOADescriptor_weird_cases(self):
        
        """
        Testing various weird cases
        """
        
        # TestCase 1: Empty List Mixer Block
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, [], p = self.p)

        # TestCase 2: Wrong Type Cost Block
        self.assertRaises(AttributeError, QAOADescriptor, [], self.mixer_hamil, self.p)
        self.assertRaises(AttributeError, QAOADescriptor, [RZGateMap(0)], 
                          self.mixer_hamil, self.p)
        
        # TestCase 3: p is not an int
        self.assertRaises(TypeError, QAOADescriptor, self.cost_hamil, 
                          self.mixer_hamil, 'test')
    
    def test_QAOADescriptor_abstract_circuit(self):
        
        """
        The abstract circuit should be consistent with the cost hamiltonian, mixer block and the p value.
        """
        
        mixer_hamil = X_mixer_hamiltonian(n_qubits = 3)
        mixer_gatemap = [RXGateMap(0), RXGateMap(1), RXGateMap(2)]
        
        qaoa_descriptor = QAOADescriptor(self.cost_hamil, self.mixer_hamil, p = self.p)
        qaoa_descriptor_gm = QAOADescriptor(self.cost_hamil, self.mixer_gatemap, 
                                              p = self.p, 
                                              mixer_coeffs = self.mixer_gatemap_coeffs)
        
        # Checks if the gates are identical
        for each_cp_gate, each_cpgm_gate in zip(qaoa_descriptor.abstract_circuit, qaoa_descriptor_gm.abstract_circuit):
            self.assertTrue(type(each_cp_gate), type(each_cpgm_gate))
            if hasattr(each_cp_gate, 'qubit_1'):
                self.assertEqual(each_cp_gate.qubit_1, each_cpgm_gate.qubit_1)
            if hasattr(each_cpgm_gate, 'qubit_2'):
                self.assertEqual(each_cp_gate.qubit_2, each_cpgm_gate.qubit_2)


class TestingQAOAVariationalParameters(unittest.TestCase):

    def test_QAOAVariationalExtendedParams(self):
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2)

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.betas_singles, [[0.75] * 3, [0.25] * 3])
        assert np.allclose(params.betas_pairs, [])
        assert np.allclose(params.gammas_singles, [[0.25], [0.75]])
        assert np.allclose(params.gammas_pairs, [[0.25, 0.25], [0.75, 0.75]])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    # TODO check that the values also make sense
    def test_QAOAVariationalExtendedParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)

        betas_singles = [[0.0, 0.1, 0.3], [0.5, 0.2, 1.2]]
        betas_pairs = []
        gammas_singles = [[0.0], [0.5]]
        gammas_pairs = [[0.1, 0.3], [0.2, 1.2]]
        params = QAOAVariationalExtendedParams(qaoa_descriptor,
                                               betas_singles,
                                               betas_pairs,
                                               gammas_singles,
                                               gammas_pairs)

        cost_2q_angles = 2 * np.array([1, -2.0]) * gammas_pairs
        cost_1q_angles = 2 * np.array([0.5]) * gammas_singles
        mixer_1q_angles = 2 * np.array([-1, -1, -1]) * betas_singles
        mixer_2q_angles = []

        assert np.allclose(params.betas_singles, betas_singles)
        assert np.allclose(params.betas_pairs, betas_pairs)
        assert np.allclose(params.gammas_singles, gammas_singles)
        assert np.allclose(params.gammas_pairs, gammas_pairs)

        assert np.allclose(params.mixer_1q_angles, mixer_1q_angles)
        assert np.allclose(params.mixer_2q_angles, mixer_2q_angles)
        assert np.allclose(params.cost_1q_angles, cost_1q_angles)
        assert np.allclose(params.cost_2q_angles, cost_2q_angles)

    def test_QAOAVariationalStandardWithBiasParamsCustomInitialisation(self):
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)

        betas = [np.pi, 0.4]
        gammas_singles = [10, 24]
        gammas_pairs = [8.8, 2.3]
        params = QAOAVariationalStandardWithBiasParams(qaoa_descriptor,
                                                       betas,
                                                       gammas_singles,
                                                       gammas_pairs)

        cost_2q_angles = 2 * np.outer(gammas_pairs, np.array([1, -2.0]))
        cost_1q_angles = 2 * np.outer(gammas_singles, np.array([0.5]))
        mixer_1q_angles = 2 * np.outer(betas, np.array([-1, -1, -1]))
        mixer_2q_angles = []

        assert np.allclose(params.betas, betas)
        assert np.allclose(params.gammas_singles, gammas_singles)
        assert np.allclose(params.gammas_pairs, gammas_pairs)

        assert np.allclose(params.mixer_1q_angles, mixer_1q_angles)
        assert np.allclose(params.mixer_2q_angles, mixer_2q_angles)
        assert np.allclose(params.cost_1q_angles, cost_1q_angles)
        assert np.allclose(params.cost_2q_angles, cost_2q_angles)
        assert type(params) == QAOAVariationalStandardWithBiasParams

    def test_QAOAVariationalAnnealingParamsCustomInitialisation(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        schedule = [0.4, 1.0]
        annealing_params = QAOAVariationalAnnealingParams(qaoa_descriptor, total_annealing_time=5,
                                                          schedule=schedule)

        assert type(annealing_params) == QAOAVariationalAnnealingParams

    def test_QAOAVariationalFourierParamsCustomInitialisation(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        v = [0.4, 1.0]
        u = [0.5, 1.2]
        fourier_params = QAOAVariationalFourierParams(
            qaoa_descriptor, q=2, v=v, u=u)

        assert np.allclose(fourier_params.betas, dct(
            v, n=qaoa_descriptor.p, axis=0))
        assert np.allclose(fourier_params.gammas, dst(
            u, n=qaoa_descriptor.p, axis=0))
        assert type(fourier_params) == QAOAVariationalFourierParams

    def test_QAOAVariationalFourierExtendedParamsCustomInitialisation(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        v_singles = [[0.4]*3, [1.0]*3]
        v_pairs = []
        u_singles = [[0.5], [1.2]]
        u_pairs = [[4.5]*2, [123]*2]
        fourier_params = QAOAVariationalFourierExtendedParams(qaoa_descriptor,
                                                              2,
                                                              v_singles,
                                                              v_pairs,
                                                              u_singles,
                                                              u_pairs)

        assert np.allclose(fourier_params.betas_singles, dct(
            v_singles, n=qaoa_descriptor.p, axis=0))
        assert np.allclose(fourier_params.gammas_singles, dst(
            u_singles, n=qaoa_descriptor.p, axis=0))
        assert np.allclose(fourier_params.gammas_pairs, dst(
            u_pairs, n=qaoa_descriptor.p, axis=0))

        assert type(fourier_params) == QAOAVariationalFourierExtendedParams

    def test_QAOAVariationalAnnealingParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalAnnealingParams.linear_ramp_from_hamiltonian(qaoa_descriptor,
                                                                             total_annealing_time=2)

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [
                           [-2*0.75] * 3, [-2*0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2*0.125], [2*0.375]])
        assert np.allclose(params.cost_2q_angles, [
                           [2*0.25, -0.5*2], [2*0.75, -1.5*2]])

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalStandardWithBiasParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalStandardWithBiasParams.linear_ramp_from_hamiltonian(qaoa_descriptor,
                                                                                    time=2)
        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [
                           [-2*0.75] * 3, [-2*0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2*0.125], [2*0.375]])
        assert np.allclose(params.cost_2q_angles, [
                           [2*0.25, -0.5*2], [2*0.75, -1.5*2]])

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalStandardParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        params = QAOAVariationalStandardParams.linear_ramp_from_hamiltonian(qaoa_descriptor,
                                                                            time=2)

        assert np.allclose(qaoa_descriptor.qureg, [0, 1, 2])
        assert np.allclose(params.mixer_1q_angles, [
                           [-2*0.75] * 3, [-2*0.25] * 3])
        assert np.allclose(params.cost_1q_angles, [[2*0.125], [2*0.375]])
        assert np.allclose(params.cost_2q_angles, [
                           [2*0.25, -0.5*2], [2*0.75, -1.5*2]])

        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_non_fourier_params_are_consistent(self):
        """
        Check that StandardParams, StandardWithBiasParams and
        ExtendedParams give the same rotation angles, given the same data"""
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        p1 = QAOAVariationalStandardParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2)
        p2 = QAOAVariationalExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2)
        p3 = QAOAVariationalStandardWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, time=2)

        assert np.allclose(p1.mixer_1q_angles, p2.mixer_1q_angles)
        assert np.allclose(p2.mixer_1q_angles, p3.mixer_1q_angles)
        assert np.allclose(p1.cost_1q_angles, p2.cost_1q_angles)
        assert np.allclose(p2.cost_1q_angles, p3.cost_1q_angles)
        assert np.allclose(p1.cost_2q_angles, p2.cost_2q_angles)
        assert np.allclose(p2.cost_2q_angles, p3.cost_2q_angles)

    def test_QAOAVariationalFourierParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2)

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        assert np.allclose(params.v, [1/3, 0])
        assert np.allclose(params.u, [1/3, 0])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalFourierWithBiasParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierWithBiasParams.linear_ramp_from_hamiltonian(qaoa_descriptor,
                                                                                   q=2,
                                                                                   time=2)

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        assert np.allclose(params.v, [1/3, 0])
        assert np.allclose(params.u_singles, [1/3, 0])
        assert np.allclose(params.u_pairs, [1/3, 0])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_QAOAVariationalFourierExtendedParams(self):

        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=3)
        params = QAOAVariationalFourierExtendedParams.linear_ramp_from_hamiltonian(qaoa_descriptor,
                                                                                   q=2,
                                                                                   time=2)

        # just access the angles, to check that it actually creates them
        assert len(params.cost_1q_angles) == len(params.cost_2q_angles)
        # Test updating and raw output

        assert np.allclose(params.v_singles, [[1/3] * 3, [0] * 3])
        assert np.allclose(params.u_singles, [[1/3] * 1, [0] * 1])
        assert np.allclose(params.u_pairs, [[1/3] * 2, [0] * 2])
        # Test updating and raw output
        raw = np.random.rand(len(params))
        params.update_from_raw(raw)
        assert np.allclose(raw, params.raw())

    def test_FourierParams_are_consistent(self):
        """
        Check, that both Fourier Parametrizations give the same rotation angles,
        given the same data"""
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=3)

        params1 = QAOAVariationalFourierParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2)
        params2 = QAOAVariationalFourierWithBiasParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2)
        params3 = QAOAVariationalFourierExtendedParams.linear_ramp_from_hamiltonian(
            qaoa_descriptor, q=2, time=2)

        assert np.allclose(params1.mixer_1q_angles, params2.mixer_1q_angles)
        assert np.allclose(params1.cost_1q_angles, params2.cost_1q_angles)
        assert np.allclose(params1.cost_2q_angles, params2.cost_2q_angles)
        assert np.allclose(params1.mixer_1q_angles, params3.mixer_1q_angles)
        assert np.allclose(params1.cost_1q_angles, params3.cost_1q_angles)
        assert np.allclose(params1.cost_2q_angles, params3.cost_2q_angles)

    def test_inputChecking(self):

        # Check that an error is raised if we pass in an extra angle in `betas` (depth is 3, here we give 4 beta values)
        reg = [0, 1]
        terms = [[0, 1]]
        weights = [0.7]
        n_layers = 3
        betas_singles = [1, 2, 3, 4]
        betas_pairs = []
        gammas_singles = []
        gammas_pairs = [1, 2, 3]
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=n_layers)
        self.assertRaises(ValueError,
                          QAOAVariationalExtendedParams,
                          qaoa_descriptor,
                          betas_singles,
                          betas_pairs,
                          gammas_singles,
                          gammas_pairs)
        
    def test_str_and_repr(self):
        
        qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p=2)
        
        for each_key_value in PARAMS_CLASSES_MAPPER.keys():
            variate_params = create_qaoa_variational_params(qaoa_descriptor, 
                                                            params_type = each_key_value, 
                                                            init_type = 'rand', 
                                                            q = 1, 
                                                            total_annealing_time = 1)
            
            self.assertEqual(variate_params.__str__(), variate_params.__repr__())
        
        
    # # Plot Tests

    # def test_StandardParams_plot(self):

    #     register_wo_bias = [0, 1, 2]
    #     terms_wo_bias = [[0, 1], [0, 2]]
    #     weights_wo_bias = [1, -2.0]

    #     p = 5
    #     hyperparams = HyperParams(terms,weights,register,p)
    #     params = StandardParams.linear_ramp_from_hamiltonian(hyperparams)
    #     fig, ax = plt.subplots()
    #     params.plot(ax=ax)
    #     # plt.show()

    #     p = 8
    #     hyperparams = HyperParams(terms,weights,register,p)
    #     params = StandardParams(hyperparams,([0.1]*p, [0.2]*p))
    #     fig, ax = plt.subplots()
    #     params.plot(ax=ax)
    #     # plt.show()

    #     p = 2
    #     hyperparams = HyperParams(terms_wo_bias,weights_wo_bias,register_wo_bias,p)
    #     params = StandardParams(hyperparams,([5]*p, [10]*p))
    #     fig, ax = plt.subplots()
    #     params.plot(ax=ax)
    #     # plt.show()

    # def test_ExtendedParams_plot(self):

    #     register_wo_bias = [0, 1, 2]
    #     terms_wo_bias = [[0, 1], [0, 2]]
    #     weights_wo_bias = [1, -2.0]

    #     p = 5
    #     hyperparams = HyperParams(terms_wo_bias,weights_wo_bias,register_wo_bias,p)
    #     params = ExtendedParams.linear_ramp_from_hamiltonian(hyperparams, p)
    #     fig, ax = plt.subplots()
    #     params.plot(ax=ax)
    #     # plt.show()

    #     p = 8
    #     hyperparams = HyperParams(terms_wo_bias,weights_wo_bias,register_wo_bias,p)
    #     params = ExtendedParams(hyperparams,
    #                             ([0.1] * p*len(register),
    #                             [],
    #                             [0.2] * p*len(weights_wo_bias)))
    #     fig, ax = plt.subplots()
    #     params.plot(ax=ax)
    #     # plt.show()

    # def test_extended_get_constraints(self):
    #     register = [0,1,2,3]
    #     terms = [[0, 1], [1, 2], [2, 3], [3, 0]]
    #     weights = [0.1, 0.3, 0.5, -0.7]
    #     p = 2

    #     hyperparams = HyperParams(terms,weights,register,p)
    #     params = ExtendedParams.linear_ramp_from_hamiltonian(hyperparams)

    #     actual_constraints = params.get_constraints()

    #     expected_constraints = [(0, 2 * np.pi), (0, 2 * np.pi),
    #                             (0, 2 * np.pi), (0, 2 * np.pi),
    #                             (0, 2 * np.pi), (0, 2 * np.pi),
    #                             (0, 2 * np.pi), (0, 2 * np.pi),
    #                             (0, 2 * np.pi / weights[0]),
    #                             (0, 2 * np.pi / weights[1]),
    #                             (0, 2 * np.pi / weights[2]),
    #                             (0, 2 * np.pi / weights[3]),
    #                             (0, 2 * np.pi / weights[0]),
    #                             (0, 2 * np.pi / weights[1]),
    #                             (0, 2 * np.pi / weights[2]),
    #                             (0, 2 * np.pi / weights[3])]

    #     assert(np.allclose(expected_constraints, actual_constraints))

    #     cost_function = QAOACostVector(params)
    #     np.random.seed(0)
    #     random_angles = np.random.uniform(-100, 100, size=len(params.raw()))
    #     value = cost_function(random_angles)

    #     normalised_angles = [random_angles[i] % actual_constraints[i][1]
    #                         for i in range(len(params.raw()))]
    #     normalised_value = cost_function(normalised_angles)

    #     assert(np.allclose(value, normalised_value))


if __name__ == "__main__":
    unittest.main()



import unittest
import json
import os
import pytest
import subprocess

from openqaoa.backends.qaoa_backend import get_qaoa_backend, DEVICE_NAME_TO_OBJECT_MAPPER, DEVICE_ACCESS_OBJECT_MAPPER
from openqaoa.qaoa_components import (
    Hamiltonian, create_qaoa_variational_params,
    QAOADescriptor
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_device import create_device
from openqaoa.backends.basebackend import QAOABaseBackendShotBased


def get_params():
    cost_hamil = Hamiltonian.classical_hamiltonian([[0, 1]], [1], constant=0)
    mixer_hamil = X_mixer_hamiltonian(2)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variational_params_std = create_qaoa_variational_params(qaoa_descriptor, 'standard', 'ramp')

    return qaoa_descriptor, variational_params_std

class TestingBackendLocal(unittest.TestCase):
    """
    These tests check that the methods of the local backends are working properly.
    """

    def test_get_counts_and_expectation_n_shots(self):
        """ 
        Check that the .get_counts admit n_shots as an argument, and works properly for the backend of all local devices.
        Also check that .expectation and .expecation_w_uncertainty methods admit n_shots as an argument for the QAOABaseBackendShotBased backends.
        """

        for device_name in DEVICE_NAME_TO_OBJECT_MAPPER.keys():
            
            # Analytical device doesn't have any of those so we are skipping it in the tests.
            if device_name in ['analytical_simulator']: continue

            qaoa_descriptor, variational_params_std= get_params()

            device = create_device(location='local', name=device_name)
            backend = get_qaoa_backend(qaoa_descriptor=qaoa_descriptor, device=device, n_shots=1000)
                        
            assert sum(backend.get_counts(params=variational_params_std, n_shots=58).values())==58, "`n_shots` is not being respected for the local simulator `{}` when calling backend.get_counts(n_shots=58).".format(device_name)
            if isinstance(backend, QAOABaseBackendShotBased): 
                try: backend.expectation(params=variational_params_std, n_shots=58)
                except Exception: raise Exception("backend.expectation does not admit `n_shots` as an argument for the local simulator `{}`.".format(device_name))
                try: backend.expectation_w_uncertainty(params=variational_params_std, n_shots=58)
                except Exception: raise Exception("backend.expectation_w_uncertainty does not admit `n_shots` as an argument for the local simulator `{}`.".format(device_name))


class TestingBackendQPUs(unittest.TestCase): 
    """ 
    These tests check methods of the QPU backends.

    For all of these tests, credentials.json MUST be filled with the appropriate
    credentials. 
    """

    @pytest.mark.qpu
    def setUp(self):

        self.HUB = 'ibm-q'
        self.GROUP = 'open'
        self.PROJECT = 'main'
        
        bashCommand = "az resource list"
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        
        if error is not None:
            print(error)
            raise Exception('You must have the Azure CLI installed and must be logged in to use the Azure Quantum Backends')
        else:
            output_json = json.loads(output)
            output_json_s = [each_json for each_json in output_json if each_json['name'] == 'TestingOpenQAOA'][0]
            self.RESOURCE_ID = output_json_s['id']
            self.AZ_LOCATION = output_json_s['location']
            

    @pytest.mark.qpu
    def test_get_counts_and_expectation_n_shots(self):
        """ Check that the .get_counts, .expectation and .expecation_w_uncertainty methods admit n_shots as an argument for the backends of all QPUs. """

        list_device_attributes = [ 
                                    {'QPU': 'Qiskit', 'device_name': 'ibmq_qasm_simulator', 'hub': self.HUB, 'group': self.GROUP, 'project': self.PROJECT}, 
                                    {'QPU': 'Pyquil', 'device_name': "2q-qvm", 'as_qvm': True, 'execution_timeout': 3, 'compiler_timeout': 3},       
                                    {'QPU': 'AWS', 'device_name': 'arn:aws:braket:::device/quantum-simulator/amazon/sv1'},
                                    {'QPU': 'Azure', 'device_name': 'rigetti.sim.qvm', 'resource_id': self.RESOURCE_ID, 'az_location': self.AZ_LOCATION}
                                ]

        assert len(list_device_attributes) == len(DEVICE_ACCESS_OBJECT_MAPPER), "The number of QPUs in the list of tests is not the same as the number of QPUs in the DEVICE_ACCESS_OBJECT_MAPPER. The list should be updated."

        for (device, backend), device_attributes in zip(DEVICE_ACCESS_OBJECT_MAPPER.items(), list_device_attributes):

            qaoa_descriptor, variational_params_std = get_params()

            QPU_name = device_attributes.pop('QPU')
            print("Testing {} backend.".format(QPU_name))


            # if QPU_name in ["AWS", 'Qiskit']:
            #     print(f"Skipping test for {QPU_name} backend.")
            #     continue 

            try: 
                device = device(**device_attributes)
                backend = backend(qaoa_descriptor = qaoa_descriptor, device = device, cvar_alpha = 1, n_shots=100, prepend_state = None, append_state = None, init_hadamard = True)

                # Check that the .get_counts, .expectation and .expectation_w_variance methods admit n_shots as an argument
                assert sum(backend.get_counts(params=variational_params_std, n_shots=58).values()) == 58, "`n_shots` is not being respected when calling .get_counts(n_shots=58).".format(QPU_name)
                backend.expectation(params=variational_params_std, n_shots=58)
                backend.expectation_w_uncertainty(params=variational_params_std, n_shots=58)

            except Exception as e: 

                raise e from type(e)(f"Error raised for `{QPU_name}`: " + str(e))

            print("Test passed for {} backend.".format(QPU_name))




if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import List, Tuple
import numpy as np

from qiskit import QuantumCircuit as qkQuantumCircuit
from qiskit.circuit.library import (
    RXGate,
    RYGate,
    RZGate,
    CXGate,
    CZGate,
    RXXGate,
    RZXGate,
    RZZGate,
    RYYGate,
    CRZGate,
)
from pyquil import Program as quilProgram
from pyquil import gates as quilgates
from pyquil.quilatom import QubitPlaceholder as quilQubitPlaceholder
from .rotationangle import RotationAngle
from braket.circuits import gates as braketgates
from braket.circuits import Circuit


class Gate(ABC):
    def __init__(self, ibm_gate, pyquil_gate, braket_gate, vector_gate):
        self.ibm_gate = ibm_gate
        self.pyquil_gate = pyquil_gate
        self.braket_gate = braket_gate
        self.vector_gate = vector_gate

    @abstractmethod
    def apply_ibm_gate(self, circuit):
        pass

    @abstractmethod
    def apply_pyquil_gate(self, circuit):
        pass

    @abstractmethod
    def apply_braket_gate(self, circuit):
        pass

    @abstractmethod
    def apply_vector_gate(self, circuit):
        pass


class OneQubitGate(Gate):
    def apply_ibm_gate(
        self,
        qubit_idx: int,
        rotation_angle_obj: RotationAngle,
        circuit: qkQuantumCircuit,
    ):
        if self.ibm_gate is not None:
            circuit.append(
                self.ibm_gate(rotation_angle_obj.rotation_angle), [qubit_idx], []
            )
        else:
            raise NotImplementedError()
        return circuit

    def apply_pyquil_gate(
        self,
        qubit_idx: quilQubitPlaceholder,
        rotation_angle_obj: RotationAngle,
        program: quilProgram,
    ):
        if self.pyquil_gate is not None:
            program += self.pyquil_gate(rotation_angle_obj.rotation_angle, qubit_idx)
        else:
            raise NotImplementedError()
        return program

    def apply_braket_gate(
        self, qubit_idx: int, rotation_angle_obj: RotationAngle, circuit: Circuit
    ):
        if self.braket_gate is not None:
            circuit += self.braket_gate(qubit_idx, rotation_angle_obj.rotation_angle)
        else:
            raise NotImplementedError()
        return circuit

    def apply_vector_gate(self, qubit_idx, rotation_angle_obj, input_obj):
        return NotImplementedError(
            "Implement this method for each supporting gate class"
        )


class RY(OneQubitGate):
    def __init__(self):

        ibm_gate = RYGate
        pyquil_gate = quilgates.RY
        braket_gate = braketgates.Ry.ry
        vector_gate = None

        super().__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_idx, rotation_angle_obj, input_obj):
        input_obj.apply_ry(qubit_idx, rotation_angle_obj.rotation_angle)


class RX(OneQubitGate):
    def __init__(self):

        ibm_gate = RXGate
        pyquil_gate = quilgates.RX
        braket_gate = braketgates.Rx.rx
        vector_gate = None

        super().__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_idx, rotation_angle_obj, input_obj):
        input_obj.apply_rx(qubit_idx, rotation_angle_obj.rotation_angle)


class RZ(OneQubitGate):
    def __init__(self):

        ibm_gate = RZGate
        pyquil_gate = quilgates.RZ
        braket_gate = braketgates.Rz.rz
        vector_gate = None

        super().__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_idx, rotation_angle_obj, input_obj):
        input_obj.apply_rz(qubit_idx, rotation_angle_obj.rotation_angle)


class TwoQubitGate(Gate):
    def apply_ibm_gate(self, qubit_indices: List[int], circuit: qkQuantumCircuit):
        if self.ibm_gate is not None:
            circuit.append(self.ibm_gate(), qubit_indices, [])
        else:
            raise NotImplementedError()
        return circuit

    def apply_pyquil_gate(
        self, qubit_indices: List[quilQubitPlaceholder], program: quilProgram
    ):
        if self.pyquil_gate is not None:
            program += self.pyquil_gate(qubit_indices[0], qubit_indices[1])
        else:
            raise NotImplementedError()
        return program

    def apply_braket_gate(self, qubit_indices: List[int], circuit: Circuit):

        if self.braket_gate is not None:
            circuit += self.braket_gate(qubit_indices[0], qubit_indices[1])
        else:
            raise NotImplementedError()
        return circuit

    def apply_vector_gate(self, qubit_indices, input_obj):
        return NotImplementedError(
            "Implement this method for each supporting gate class"
        )


class CZ(TwoQubitGate):
    def __init__(self):

        ibm_gate = CZGate
        pyquil_gate = quilgates.CZ
        braket_gate = braketgates.CZ.cz
        vector_gate = None

        super().__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self):

        raise NotImplementedError("This gate is not yet supported")


class CX(TwoQubitGate):
    def __init__(self, mode: str = "CX"):

        ibm_gate = CXGate
        pyquil_gate = quilgates.CNOT
        braket_gate = braketgates.CNot.cnot
        vector_gate = None

        super().__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )
        self.mode = mode

    def _in_XY(self, qubit_indices) -> List[Tuple[Gate, List]]:
        qubit_1 = qubit_indices[0]
        qubit_2 = qubit_indices[1]
        return [
            (RX, [qubit_2, RotationAngle(lambda x: x, [], np.pi / 2)]),
            (RZ, [qubit_1, RotationAngle(lambda x: x, [], -np.pi / 2)]),
            (RZ, [qubit_1, RotationAngle(lambda x: x, [], np.pi / 2)]),
            (RiSWAP, [[qubit_1, qubit_2, RotationAngle(lambda x: x, [], np.pi)]]),
            (RX, [qubit_1, RotationAngle(lambda x: x, [], np.pi / 2)]),
            (RiSWAP, [[qubit_1, qubit_2, RotationAngle(lambda x: x, [], np.pi)]]),
            (RZ, [qubit_2, RotationAngle(lambda x: x, [], np.pi / 2)]),
        ]

    def _in_CZ(self, qubit_indices) -> List[Tuple[Gate, List]]:
        qubit_1 = qubit_indices[0]
        qubit_2 = qubit_indices[1]
        return [
            (RY, [qubit_2, RotationAngle(lambda x: x, [], np.pi / 2)]),
            (RX, [qubit_2, RotationAngle(lambda x: x, [], np.pi)]),
            (CZ, [[qubit_1, qubit_2]]),
            (RY, [qubit_2, RotationAngle(lambda x: x, [], np.pi / 2)]),
            (RX, [qubit_2, RotationAngle(lambda x: x, [], np.pi)]),
        ]

    def apply_ibm_gate(
        self, qubit_indices: List[int], circuit: qkQuantumCircuit
    ) -> qkQuantumCircuit:

        if self.mode == "CX":
            circuit.cx(qubit_indices[0], qubit_indices[1])
        elif self.mode == "CZ":
            for each_object, init_params in self._in_CZ(qubit_indices):
                circuit = each_object().apply_ibm_gate(*init_params, circuit)

        return circuit

    def apply_vector_gate(self, input_obj):

        raise NotImplementedError("This gate is not yet supported")


class TwoQubitGateWithAngle(TwoQubitGate):
    def apply_ibm_gate(
        self,
        qubit_indices: List[int],
        rotation_angle_obj: RotationAngle,
        circuit: qkQuantumCircuit,
    ):
        if self.ibm_gate is not None:
            circuit.append(
                self.ibm_gate(rotation_angle_obj.rotation_angle), qubit_indices, []
            )
        else:
            raise NotImplementedError()
        return circuit

    def apply_pyquil_gate(
        self,
        qubit_indices: List[quilQubitPlaceholder],
        rotation_angle_obj: RotationAngle,
        program: quilProgram,
    ):
        if self.pyquil_gate is not None:
            program += self.pyquil_gate(
                rotation_angle_obj.rotation_angle, qubit_indices[0], qubit_indices[1]
            )
        else:
            raise NotImplementedError()
        return program

    def apply_braket_gate(self, qubit_indices, rotation_angle_obj, circuit):

        if self.braket_gate is not None:
            circuit += self.braket_gate(
                qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
            )
        else:
            raise NotImplementedError()
        return circuit

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        return NotImplementedError(
            "Implement this method for each supporting gate class"
        )


class RXX(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = RXXGate
        pyquil_gate = None
        braket_gate = braketgates.XX.xx
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_rxx(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class RYY(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = RYYGate
        pyquil_gate = None
        braket_gate = braketgates.YY.yy
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_ryy(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class RZZ(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = RZZGate
        pyquil_gate = None
        braket_gate = braketgates.ZZ.zz
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_rzz(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class RXY(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = None
        pyquil_gate = None
        braket_gate = None
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_rxy(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class RZX(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = RZXGate
        pyquil_gate = None
        braket_gate = None
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_ibm_gate(
        self,
        qubit_indices: List[int],
        rotation_angle_obj: RotationAngle,
        circuit: qkQuantumCircuit,
    ):

        circuit.rzx(
            rotation_angle_obj.rotation_angle, qubit_indices[0], qubit_indices[1]
        )
        return circuit

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_rzx(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class RYZ(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = None
        pyquil_gate = None
        braket_gate = None
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def apply_vector_gate(self, qubit_indices, rotation_angle_obj, input_obj):
        input_obj.apply_ryz(
            qubit_indices[0], qubit_indices[1], rotation_angle_obj.rotation_angle
        )


class CPHASE(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = CRZGate
        pyquil_gate = quilgates.CPHASE
        braket_gate = braketgates.CPhaseShift.cphaseshift
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def _vector_gate(self):

        raise NotImplementedError("This gate is not yet supported")


class RiSWAP(TwoQubitGateWithAngle):
    def __init__(self):

        ibm_gate = None
        pyquil_gate = quilgates.XY
        braket_gate = braketgates.XY.xy
        vector_gate = None

        super(TwoQubitGate, self).__init__(
            ibm_gate=ibm_gate,
            pyquil_gate=pyquil_gate,
            braket_gate=braket_gate,
            vector_gate=vector_gate,
        )

    def _vector_gate(self):

        raise NotImplementedError("This gate is not yet supported")



from __future__ import annotations

from .devices_core import DeviceBase, DeviceLocal
from openqaoa_braket.backends import DeviceAWS
from openqaoa_qiskit.backends import DeviceQiskit
from openqaoa_pyquil.backends import DevicePyquil
from openqaoa_azure.backends import DeviceAzure


def device_class_arg_mapper(
    device_class: DeviceBase,
    hub: str = None,
    group: str = None,
    project: str = None,
    as_qvm: bool = None,
    noisy: bool = None,
    compiler_timeout: float = None,
    execution_timeout: float = None,
    client_configuration: QCSClientConfiguration = None,
    endpoint_id: str = None,
    engagement_manager: EngagementManager = None,
    folder_name: str = None,
    s3_bucket_name: str = None,
    aws_region: str = None,
    resource_id: str = None,
    az_location: str = None,
) -> dict:
    DEVICE_ARGS_MAPPER = {
        DeviceQiskit: {"hub": hub, "group": group, "project": project},
        DevicePyquil: {
            "as_qvm": as_qvm,
            "noisy": noisy,
            "compiler_timeout": compiler_timeout,
            "execution_timeout": execution_timeout,
            "client_configuration": client_configuration,
            "endpoint_id": endpoint_id,
            "engagement_manager": engagement_manager,
        },
        DeviceAWS: {
            "s3_bucket_name": s3_bucket_name,
            "aws_region": aws_region,
            "folder_name": folder_name,
        },
        DeviceAzure: {"resource_id": resource_id, "location": az_location},
    }

    final_device_kwargs = {
        key: value
        for key, value in DEVICE_ARGS_MAPPER[device_class].items()
        if value is not None
    }
    return final_device_kwargs


def create_device(location: str, name: str, **kwargs):
    """
    This function returns an instance of the appropriate device class.

    Parameters
    ----------
    device_name: str
        The name of the device to be accessed.
    device_location: str
        The location of the device to be accessed.
    kwargs: dict
        A dictionary of keyword arguments to be passed to the device class.
        These will be used to initialise the device.

    Returns
    -------
    device: DeviceBase
        An instance of the appropriate device class.
    """
    location = location.lower()
    if location == "ibmq":
        device_class = DeviceQiskit
    elif location == "qcs":
        device_class = DevicePyquil
    elif location == "aws":
        device_class = DeviceAWS
    elif location == "local":
        device_class = DeviceLocal
    elif location == "azure":
        device_class = DeviceAzure
    else:
        raise ValueError(f"Invalid device location, Choose from: {location}")

    return device_class(device_name=name, **kwargs)



from typing import List, Union, Callable


class RotationAngle(object):
    def __init__(
        self,
        angle_relationship: Callable,
        pauli_label: List,
        pauli_angle: Union[int, float] = None,
    ):

        self._angle = angle_relationship
        self.pauli_label = pauli_label
        self.pauli_angle = pauli_angle

    @property
    def rotation_angle(self) -> Union[int, float]:

        return self._angle(self.pauli_angle)



import warnings
import numpy as np
import unittest

# OpenQAOA imports
from openqaoa.backends import QAOAvectorizedBackendSimulator
from openqaoa.qaoa_components import QAOADescriptor, Hamiltonian, create_qaoa_variational_params
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.derivatives.derivative_functions import derivative

"""
Unittest based testing of derivative computations.
"""


class TestQAOACostBaseClass(unittest.TestCase):

    '''
    def test_gradient_agreement(self):
        "Test agreement between gradients computed from finite difference, parameter shift and SPS (all gates sampled) for weighted and unweighted graphs at several parameters."
        
        # unweighted graph
        terms = [[0,1], [0,2], [1,3], [2]]
        weights = [1, 1, 1, 1]
        register = [0, 1, 2, 3]
        p = 2
        nqubits = 4

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(
            qaoa_qaoa_descriptor, 'standard', 'ramp')
        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True)
        
        grad_stepsize = 0.00000001
        gradient_ps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'param_shift')
        gradient_fd = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'finite_difference', {'stepsize': grad_stepsize})
        gradient_sps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'stoch_param_shift', {'stepsize':grad_stepsize, 'n_beta_single':-1, 'n_beta_pair':-1, 'n_gamma_pair':-1, 'n_gamma_single':-1})

        params = [[0,0,0,0], [1,1,1,1], [np.pi/2, np.pi/2, np.pi/2, np.pi/2]]

        for param in params:
            grad_fd = gradient_fd(param)
            grad_ps = gradient_ps(param)
            grad_sps = gradient_sps(param)

            for i, grad in enumerate(grad_fd): 
                assert np.isclose(grad, grad_ps[i], rtol=1e-05, atol=1e-05)
                assert np.isclose(grad, grad_sps[i], rtol=1e-05, atol=1e-05)

        # weighted graph with bias
        terms = [[0,1], [1,2], [0,3], [2], [1]]
        weights = [1, 1.1, 1.5, 2, -0.8]
        register = [0, 1, 2, 3]
        p = 2
        nqubits = 4

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms, weights, constant=0.8)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(
            cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(
            qaoa_qaoa_descriptor, 'standard', 'ramp')
        backend_vectorized = QAOAvectorizedBackendSimulator(
            qaoa_qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True)
        
        grad_stepsize = 0.00000001
        gradient_ps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'param_shift')
        gradient_fd = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'finite_difference', {'stepsize': grad_stepsize})
        gradient_sps = backend_vectorized.derivative_function(variational_params_std, 'gradient', 'stoch_param_shift', {'stepsize':grad_stepsize, 'n_beta_single':-1, 'n_beta_pair':-1, 'n_gamma_pair':-1, 'n_gamma_single':-1})
    
        params = [[0,0,0,0], [1,1,1,1], [np.pi/2, np.pi/2, np.pi/2, np.pi/2]]

        for param in params:
            grad_fd = gradient_fd(param)
            grad_ps = gradient_ps(param)
            grad_sps = gradient_sps(param)

            for i, grad in enumerate(grad_fd): 
                assert np.isclose(grad, grad_ps[i], rtol=1e-05, atol=1e-05)
                assert np.isclose(grad, grad_sps[i], rtol=1e-05, atol=1e-05)
    '''
    
    def setUp(self):
        
        self.log = Logger({'func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           }, 
                           'jac_func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           },
                           'n_shots':
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'Replace'
                            }
                          }, 
                          {
                              'root_nodes': ['func_evals', 'jac_func_evals', 'n_shots'],
                              'best_update_structure': []
                          })
        
        self.log.log_variables({'func_evals': 0})
        self.log.log_variables({'jac_func_evals': 0})

    def __backend_params(self, terms:list, weights:list, p:int, nqubits:int):
        cost_hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(nqubits)
        qaoa_qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = create_qaoa_variational_params(qaoa_qaoa_descriptor, 'standard', 'ramp')
        backend_vectorized = QAOAvectorizedBackendSimulator(qaoa_qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True)
        return backend_vectorized, variational_params_std

    def test_gradient_computation(self):
        "Test gradient computation by param. shift, finite difference, and SPS (all gates sampled) on barbell graph."

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(terms=[[0, 1]], weights=[1], p=1, nqubits=2)
        
        gradients_types_list = ['finite_difference', 'param_shift', 'stoch_param_shift']
        gradients_fun_list = [derivative(backend, params, self.log, 'gradient', type_, {'stepsize': 0.0000001}) for type_ in gradients_types_list]

        test_points = [[0, 0], [np.pi/2, np.pi/3], [1, 2]]

        for point in test_points:
            beta, gamma = point

            dCdb, dCdg = -4*np.cos(4*beta)*np.sin(2*gamma), -2*np.sin(4*beta)*np.cos(2*gamma)

            for gradient_fun, gradient_name in zip(gradients_fun_list, gradients_types_list):
                dCdb_, dCdg_ = gradient_fun(point)
                assert np.isclose(dCdb, dCdb_, rtol=1e-05, atol=1e-05), f'Gradient computation failed for {gradient_name} on barbell graph. dCdb: {dCdb}, dCdb_: {dCdb_}'
                assert np.isclose(dCdg, dCdg_, rtol=1e-05, atol=1e-05), f'Gradient computation failed for {gradient_name} on barbell graph. dCdg: {dCdg}, dCdg_: {dCdg_}'

    def test_gradient_w_variance_computation(self):
        "Test gradient computation by param. shift, finite difference, and SPS (all gates sampled) on barbell graph."

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(terms=[[0, 1]], weights=[1], p=1, nqubits=2)

        gradients_types_list = ['finite_difference', 'param_shift', 'stoch_param_shift']
        gradients_fun_list = [derivative(backend, params, self.log, 'gradient_w_variance', type_, {'stepsize': 0.0000001}) for type_ in gradients_types_list]

        test_points = [[0, 0], [np.pi/2, np.pi/3], [1, 2]]

        for point in test_points: 

            for gradient_fun, gradient_name in zip(gradients_fun_list, gradients_types_list):

                #compute gradient for each point with number of shots 1000 for ech function evaluation 
                grad, var, n_shots = gradient_fun(point, n_shots=1000)

                #check if there is a gradient and variance (we can't check the value of the gradient and variance because it is randomly computed, since n_shots is 1000)
                for g in grad: assert np.abs(g)>=0, f'Gradient computation failed for {gradient_name} on barbell graph. grad: {grad}'
                for v in var: assert v>0, f'Error computing the variance of the gradient for {gradient_name} on barbell graph.'

                #check if the number of shots is correct
                x = 4 if gradient_name == 'finite_difference' else 6 #that is also checking that  SPS samples all gates when (n_beta, n_gamma_pair, n_gamma_single) is (-1, -1, -1)
                assert n_shots == x*1000, f'The number of shots should be {x*1000} but is {n_shots}.'

    def test_SPS_sampling(self):
        "Test that SPS samples the number of gates specified by the user."
        backend, params = self.__backend_params(terms=[[0, 1]], weights=[1], p=1, nqubits=2)
        gradient_fun = derivative( backend, params, self.log, 'gradient_w_variance', 
                                    'stoch_param_shift', {'stepsize':0.1, 'n_beta_single':1, 'n_beta_pair':0, 'n_gamma_pair':1, 'n_gamma_single':0})

        n_shots = gradient_fun([0,0], n_shots=1000)[2]
        #check if the number of shots is correct, we know that we are sampling 2 gates (1 beta and 1 gamma) -> 2*2*1000 (the second 2* comes from two evaluations for each gradient)
        assert n_shots == 4000, f'The number of shots should be {4000} but is {n_shots}.'

    
    def test_hessian_computation(self):
        "Test Hessian computation by finite difference on barbell graph"

        # Analytical cost expression : C(b,g) = -sin(4b)*sin(2g)
        backend, params = self.__backend_params(terms=[[0, 1]], weights=[1], p=1, nqubits=2)

        hessian_fd = derivative(backend, params, 
                                self.log,'hessian', 'finite_difference', 
                                {'stepsize': 0.001})

        test_points = np.round([[0, 0], [np.pi/2, np.pi/3], [1, 2]], 12)

        for point in test_points:
            beta, gamma = point[0], point[1]

            dCdbb = 16*np.sin(4*beta)*np.sin(2*gamma)
            dCdbg = -8*np.cos(4*beta)*np.cos(2*gamma)
            dCdgb = dCdbg
            dCdgg = 4*np.sin(4*beta)*np.sin(2*gamma)

            assert np.isclose(dCdbb, hessian_fd(
                point)[0][0], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdbg, hessian_fd(
                point)[0][1], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdgb, hessian_fd(
                point)[1][0], rtol=1e-05, atol=1e-05)
            assert np.isclose(dCdgg, hessian_fd(
                point)[1][1], rtol=1e-05, atol=1e-05)



if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=PendingDeprecationWarning)
        unittest.main()



from typing import List
from copy import deepcopy

from .gatemap import RotationGateMap, RotationGateMapFactory
from .operators import Hamiltonian


class HamiltonianMapper(object):
    def get_gate_maps(
        hamil_obj: Hamiltonian, input_label: List = []
    ) -> List[RotationGateMap]:
        """
        This method gets the pauli gates based on the input Hamiltonian into the Mapper

        Parameters
        ----------
        hamil_obj : `Hamiltonian`
            The Hamiltonian object to construct the gates from
        input_label : `list`
            Input label defining the type of gate

        Return
        ------
        `list[RotationGateMap]`
            List of RotationGateMap objects defining part of the circuit
        """

        assert type(input_label) is list, "input_label must be of type list"

        return RotationGateMapFactory.convert_hamiltonian_to_gate_maps(
            hamil_obj, input_label
        )

    def repeat_gate_maps(
        hamil_obj: Hamiltonian, tag: str, n_repetitions: int
    ) -> List[List[RotationGateMap]]:
        """
        Repeat the gates for n_repetitions layers based on the input Hamiltonian into the Mapper.

        Parameters
        ----------
        hamil_obj : `Hamiltonian`
            The Hamiltonian object to construct the gates from
        tag : `str`
            The tag to be used for the repeated gates
        n_repetitions: `int`
            The number of times the layer of gates have to be repeated.
        """
        output_gate_list = []

        for each_repetition in range(n_repetitions):
            output_gate_list.append(
                HamiltonianMapper.get_gate_maps(
                    hamil_obj, [tag.lower(), each_repetition]
                )
            )

        return output_gate_list

    def remap_gate_map_labels(
        gatemap_list: List[RotationGateMap], input_label: List = []
    ) -> List[RotationGateMap]:

        """
        This method reassigns the pauli_label of all the gates in the
        RotationGateMap list. The newly assigned labels help the circuit
        identify which variational angles fit into which gate.

        Parameters
        ----------
        gatemap_list: `list[RotationGateMap]`
            The list of RotationGateMap objects that needs the pauli_label attribute to be remapped.
        input_label : `list`
            Input label defining the type of gate

        Return
        ------
        `list[RotationGateMap]`
            List of RotationGateMap objects defining part of the circuit
        """

        assert type(input_label) is list, "input_label must be of type list"

        return RotationGateMapFactory.remap_gate_map_labels(gatemap_list, input_label)

    def repeat_gate_maps_from_gate_map_list(
        gatemap_list: List[RotationGateMap], tag: str, n_repetitions: int
    ) -> List[List[RotationGateMap]]:

        """
        Repeat the gates for n_repetitions layers based on the input list of RotationGateMap objects.

        Parameters
        ----------
        gatemap_list: `list[RotationGateMap]`
            The list of RotationGateMap objects that needs to be cloned.
        tag : `str`
            The tag to be used for the repeated gates
        n_repetitions: `int`
            The number of times to clone the RotationGateMap objects in the list.
        """

        output_gate_list = []

        for each_repetition in range(n_repetitions):
            output_gate_list.append(
                deepcopy(
                    HamiltonianMapper.remap_gate_map_labels(
                        gatemap_list, [tag.lower(), each_repetition]
                    )
                )
            )

        return output_gate_list



import warnings
import unittest
import numpy as np
import networkx as nx
import copy
import inspect

from openqaoa import QAOA
import openqaoa.optimizers.pennylane as pl
from openqaoa.backends import create_device
from openqaoa.optimizers.training_vqa import PennyLaneOptimizer
from openqaoa.optimizers.pennylane.optimization_methods_pennylane import AVAILABLE_OPTIMIZERS
from openqaoa.derivatives.derivative_functions import derivative
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.qaoa_components import create_qaoa_variational_params, QAOADescriptor, PauliOp, Hamiltonian
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.optimizers import get_optimizer
from openqaoa.derivatives.qfim import qfim as Qfim
from openqaoa.problems import QUBO, MinimumVertexCover


#list of optimizers to test, pennylane optimizers
list_optimizers = PennyLaneOptimizer.PENNYLANE_OPTIMIZERS

#create a problem
g = nx.circulant_graph(4, [1])
problem = MinimumVertexCover(g, field =1.0, penalty=10)
qubo_problem_1 = problem.qubo
qubo_problem_2 = QUBO.random_instance(5)
qubo_problem_3 = QUBO.random_instance(6)


class TestPennylaneOptimizers(unittest.TestCase):

    def setUp(self):
        
        self.log = Logger({'func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           },
                           'jac_func_evals':
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           },
                           'qfim_func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           }
                          }, 
                          {
                              'root_nodes': ['func_evals', 'jac_func_evals', 
                                             'qfim_func_evals'], 
                              'best_update_structure': []
                          })
        
        self.log.log_variables({'func_evals': 0, 'jac_func_evals': 0, 'qfim_func_evals': 0})

    def _run_method_workflows(self, method, problem):
        " helper function to run the test for any method using workflows"
        q = QAOA()
        q.set_classical_optimizer(method=method, maxiter=3, jac='finite_difference')
        q.compile(problem) 
        q.optimize()

        assert len(q.result.most_probable_states['solutions_bitstrings'][0]) > 0

    def _run_method_manual(self, method, problem):
        " helper function tu run the test for any method using manual mode"

        cost_hamil = problem.hamiltonian
        mixer_hamil = X_mixer_hamiltonian(n_qubits=problem.n)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device('local','vectorized')
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor,device)
        variate_params = create_qaoa_variational_params(qaoa_descriptor, 'standard', 'ramp')
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        # declare needed functions
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                    'gradient', 'finite_difference', 
                    {'stepsize': grad_stepsize})
        qfim = Qfim(backend_obj_vectorized, variate_params, self.log)   


        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                        'method': method, 'jac': jac, 'maxiter': niter, 'qfim': qfim,
                                        'optimizer_options' : {'stepsize': stepsize}})
        vector_optimizer()

        # saving the results
        results = vector_optimizer.qaoa_result

        assert len(results.most_probable_states['solutions_bitstrings'][0]) == problem.n

    def test_pennylane_optimizers_workflows(self):
        " function to run the tests for pennylane optimizers, workflows"

        i = 0
        for problem in [qubo_problem_3, qubo_problem_2, qubo_problem_1]:
            for opt in list_optimizers:
                self._run_method_workflows(opt, problem)
                i += 1

        assert i == 3*len(list_optimizers)

    def test_pennylane_optimizers_manual(self):
        " function to run the tests for pennylane optimizers, manual mode"

        i = 0
        for problem in [qubo_problem_3, qubo_problem_2, qubo_problem_1]:
            for opt in list_optimizers:
                self._run_method_manual(opt, problem)
                i += 1

        assert i == 3*len(list_optimizers)

    def _pennylane_step(self, params_array, cost, optimizer, method, jac, qfim):
        " helper function to run a setp of the pennylane optimizer"
        params_array = pl.numpy.array(params_array, requires_grad=True)
        if method in ['pennylane_adagrad', 'pennylane_adam', 'pennylane_vgd', 'pennylane_momentum', 'pennylane_nesterov_momentum', 'pennylane_rmsprop']:
            x, y = optimizer.step_and_cost(cost, params_array, grad_fn=jac)
        if method in ['pennylane_rotosolve']: 
            x, y = optimizer.step_and_cost(
                                            cost, params_array,
                                            nums_frequency={'params': {(i,):1 for i in range(params_array.size)}},
                                           )
        if method in ['pennylane_spsa']:       
            x, y = optimizer.step_and_cost(cost, params_array)

        return x, y

    def test_step_and_cost(self):
        " function to run the tests for steps of pennylane optimizers "

        # define some problem
        cost_hamil = Hamiltonian([PauliOp('ZZ', (0, 1)), PauliOp('ZZ', (1, 2)), PauliOp(
            'ZZ', (0, 3)), PauliOp('Z', (2,)), PauliOp('Z', (1,))], [1, 1.1, 1.5, 2, -0.8], 0.8)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=4)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device('local','vectorized')
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor,device)
        variate_params = create_qaoa_variational_params(qaoa_descriptor, 'standard', 'ramp')
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        # declare needed functions
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                    'gradient', 'finite_difference', 
                    {'stepsize': grad_stepsize})
        qfim = Qfim(backend_obj_vectorized, variate_params, self.log)   
        def cost(params):
            variate_params.update_from_raw(params)
            return np.real(backend_obj_vectorized.expectation(variate_params))

        i = 0
        for method in list_optimizers:
        
            pennylane_method = method

            # copy the parameters
            x0 = copy.deepcopy(variate_params.raw().copy())

            # Optimize with the implemented optimizer in OpenQAOA
            vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                            'method': method, 'jac': jac, 'maxiter': niter, 'qfim': qfim,
                                            'optimizer_options' : {'stepsize': stepsize}})
            vector_optimizer()

            # formatting the data
            y_opt = vector_optimizer.qaoa_result.intermediate['cost'][1:4]
            if pennylane_method in ['pennylane_rotosolve']: y_opt = vector_optimizer.qaoa_result.intermediate['cost'][4:40:12]

            # get optimizer to try
            optimizer = AVAILABLE_OPTIMIZERS[pennylane_method]
            #get optimizer arguments
            arguments = inspect.signature(optimizer).parameters.keys()

            #check if stepsize is in the optimizer arguments
            options = {}
            if 'stepsize' in arguments: options['stepsize'] = stepsize
            if 'maxiter'  in arguments: options['maxiter'] = niter

            #pass the argument to the optimizer
            optimizer = optimizer(**options) 

            # reinitialize variables
            variate_params.update_from_raw(x0)
            x0 = variate_params.raw().copy()
            y0 = cost(x0)

            # compute steps (depends on the optimizer)
            x1, y1 = self._pennylane_step(x0, cost, optimizer, pennylane_method, jac, qfim)
            x2, y2 = self._pennylane_step(x1, cost, optimizer, pennylane_method, jac, qfim)
            x3, y3 = self._pennylane_step(x2, cost, optimizer, pennylane_method, jac, qfim)

            # list of results
            y = [y1, y2, y3]

            # check that the results are ok
            if pennylane_method in ['pennylane_spsa']: 
                assert np.sum(np.abs(np.array(y)) >= 0) == 3
            else:
                for yi, y_opt_i in zip(y, y_opt):
                    assert np.isclose(yi, y_opt_i, rtol=0.001, atol=0.001)

            i += 1                

        assert i == len(list_optimizers)


if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=PendingDeprecationWarning)
        unittest.main()



import numpy as np
from sympy import Symbol
import unittest

from openqaoa.qaoa_components import PauliOp, Hamiltonian
"""
Unittest based testing of the PauliOp and Hamiltonian classes
"""

class TestingOperators(unittest.TestCase):

    def test_pauli_init(self):
        """
        Test the initialization method of the PauliOp class.

        The test consists in checking correct initializations for representative examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 1), (4, 5, 0, 1), (2, 0, 7, 4, 5),
                         (0, 1, 5, 3, 7, 6), (6, 5, 4, 3, 1, 1, 0)]
        input_strings = ['XZX', 'YYYY', 'ZYXZX', 'YYZZXX', 'ZYXYZYX']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correctly ordered indices and strings
        correct_indices = [(0, 1), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                           (0, 1, 3, 5, 6, 7), (0, 1, 3, 4, 5, 6)]
        correct_strings = ['XY', 'YYYY', 'YZZXX', 'YYZZXX', 'XXYXYZ']
        correct_phases = [1j,1,1,1,-1j]

        # Extract indices and strings generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]
        
        # Test that indices and strings were generated correctly
        assert indices == correct_indices, f'Pauli indices not generated correctly'
        assert strings == correct_strings, f'Pauli strings not generated correctly'
        assert np.allclose(phases,correct_phases), f'Pauli phases not generated correctly'


        # Exception cases - not valid Pauli operators
        input_exc_indices = [(0, 2, 1), (1,), (2, 2, 2, 4, 5), (0, 1, 5, 3, 7, 6), (6, 5, 4, 3, 2, 1, 0)]
        input_exc_strings = ['XY', 'XZZYZXYZ', 'YXZZ', 'XYZZX', 'Z']

        # Test the exceptions are raised
        for string,index in zip(input_exc_strings, input_exc_indices):
            
            # Attempt construction of Pauli operator 
            with self.assertRaises(AssertionError) as context:
                pauli = PauliOp(string, index)

            # Check exception message
            self.assertEqual(f"Each Pauli operator must have a unique qubit index", str(context.exception))

    def test_sort_pauli_op(self):
        """
        Test of the function that sorts the string attribute and the indices
        of the pauli operator object.

        The test consists in properly ordering a set of string + indices.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 2, 1), (4, 5, 0, 1), (2, 0, 7, 4, 5),
                         (0, 1, 5, 3, 7, 6), (6, 5, 4, 3, 2, 1, 0)]
        input_strings = ['XXZ', 'YYYY', 'ZYXZX', 'YYZZXX', 'ZYXYZYX']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correctly ordered indices and strings
        correct_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                           (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        correct_strings = ['XZX', 'YYYY', 'YZZXX', 'YYZZXX', 'XYZYXYZ']

        # Extract indices and strings generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert indices == correct_indices, f'Pauli indices not sorted correctly'
        assert strings == correct_strings, f'Pauli strings not sorted correctly'


        # Exception cases - not valid Pauli operators
        input_exc_indices = [(0, 2, 1), (1,), (2, 0, 7, 4, 5), (0, 1, 5, 3, 7, 6), (6, 5, 4, 3, 2, 1, 0)]
        input_exc_strings = ['LXY', 'Q', 'ZYXZR', 'FYZZXX', 'ZYXYZYN']

        # Pauli set
        PAULIS_SET = set('XYZI')

        # Test the exceptions are raised
        for string,index in zip(input_exc_strings, input_exc_indices):
            
            # Extract exception from string
            non_pauli_operator = list(set(string) - PAULIS_SET)[0]

            # Attempt construction of Pauli operator 
            with self.assertRaises(ValueError) as context:
                pauli = PauliOp(string, index)

            # Check exception message
            self.assertEqual(f"{non_pauli_operator} is not a valid Pauli. Please choose from the set {PAULIS_SET}", str(context.exception))

    def test_pauli_simplify(self):
        """
        Test of the function that simplifies the expression of Pauli Operator by using
        the multiplication rules that Pauli matrices obey.

        The test consists in simplify a set of Pauli operators.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 1), (0, 0, 0, 2),
                         (0, 1, 3, 3, 4), (1, 1, 1, 1, 1, 1, 1)]
        input_strings = ['ZXY', 'YYYY', 'ZYIZX', 'ZZZZZXZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Indices, strings and phases from correctly simplified Pauli operators
        correct_indices = [(0, 1), (0, 2), (0, 1, 3, 4), (1,)]
        correct_strings = ['ZZ', 'YY', 'ZYZX', 'X']
        correct_phases = [1j, 1, 1, -1]

        # Extract indices, strings and phases generated by the Pauli operators
        indices = [pauli.qubit_indices for pauli in paulis]
        strings = [pauli.pauli_str for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert indices == correct_indices, f'Indices not simplified correctly'
        assert strings == correct_strings, f'Strings not simplified correctly'
        assert np.allclose(
            phases, correct_phases), f'Phases not simplified correctly'

    def test_pauli_is_trivial(self):
        """
        Test of the function that checks if a Pauli Operator object is trivial,
        i.e. only contains identity operators.

        The tests consists in checking the triviality of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['III', 'IYYY', 'YZIXX', 'IIIIII', 'XIIIIII']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correct list containing whether the Pauli Operator is trivial or not
        correct_triviality = [True, False, False, True, False]

        # Compute triviality of the Pauli operators
        triviality = [pauli._is_trivial for pauli in paulis]

        # Test that indices and strings were correctly ordered when creating the Pauli Operator
        assert np.allclose(
            triviality, correct_triviality), f'Triviality classification is incorrect'

    def test_pauli_matrix(self):
        """
        Test of the function that computes the matrix representation of the Pauli operator.

        The test consists in computing the matrix representation of set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1), (0, 1), (0, 1), (0, 1), (0, 1, 2)]
        input_strings = ['IX', 'XY', 'XZ', 'ZY', 'IIZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correct pauli representations
        correct_matrix_representations = [np.array(([0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]), dtype=complex),
                                          np.array(
                                              ([0, 0, 0, -1j], [0, 0, 1j, 0], [0, -1j, 0, 0], [1j, 0, 0, 0]), dtype=complex),
                                          np.array(
                                              ([0, 0, 1, 0], [0, 0, 0, -1], [1, 0, 0, 0], [0, -1, 0, 0]), dtype=complex),
                                          np.array(([0, -1j, 0, 0], [1j, 0, 0, 0], [0, 0, 0, 1j],
                                                    [0, 0, -1j, 0]), dtype=complex),
                                          np.array(([1, 0, 0, 0, 0, 0, 0, 0], [0, -1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, -1, 0, 0, 0, 0],
                                                    [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, -1, 0, 0], [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, -1]), dtype=complex)]

        # Compute matrix representations
        matrix_representations = [pauli.matrix for pauli in paulis]

        # Compare each matrix representation
        matrix_comparison = [(correct_matrix_representations[i] == matrix_representations[i]).all(
        ) for i in range(len(matrix_representations))]

        # Test that the computed representations are correct
        assert np.all(matrix_comparison), f'Matrix representation is incorrect'

    def test_pauli_len(self):
        """
        Test the function that computes the length of the Pauli operator.

        The tests consists in computing the length for several examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['III', 'IYYY', 'YZIXX', 'IIIIII', 'XIIIIII']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correct Pauli operator lengths
        correct_lengths = [3, 4, 5, 6, 7]

        # Compute Pauli operator lengths
        lengths = [pauli.__len__() for pauli in paulis]

        # Test that computed lengths are correct
        assert np.allclose(
            lengths, correct_lengths), f'Computed lengths are incorrect'

    def test_pauli_eq(self):
        """
        Test the function that checks whether two Pauli operators are equivalent.

        The test consists in checking equivalence between different Pauli operators for several examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['XZX', 'XZX', 'YYYY',
                         'YZZXX', 'YYZZXX', 'YYZZXX', 'XYZYXYZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Check equivalence between contiguous Pauli operators in the list
        equivalences = [paulis[i].__eq__(paulis[i+1])
                        for i in range(len(paulis)-1)]

        # Correct equivalences
        correct_equivalences = [True, False, False, False, True, False]

        # Test the checks have been performes correctly
        assert np.allclose(
            equivalences, correct_equivalences), f'Equivalences have not been computed correctly'

    def test_pauli_copy(self):
        """
        Test the function that computes a copy of a Pauli operator.

        The test consists in creating copies for a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['XZX', 'YYYY', 'YZZXX', 'YYZZXX', 'XYZYXYZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Generate copies of the Pauli operators
        copy_paulis = [pauli.__copy__() for pauli in paulis]

        # Test the copies have been correctly performes

        check_copy = [paulis[j].__eq__(copy_paulis[j])
                      for j in range(len(paulis))]

        assert sum(check_copy) == len(
            check_copy), f'Copies were not performed correctly'

    def test_pauli_str(self):
        """
        Test the function that generates a string representation of the Pauli operator.

        The test consists in computing the string representation of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['XZX', 'YYYY', 'YZZXX', 'YYZZXX', 'XYZYXYZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correct string representations
        correct_str_rep = ['X_{0}Z_{1}X_{2}', 'Y_{0}Y_{1}Y_{4}Y_{5}', 'Y_{0}Z_{2}Z_{4}X_{5}X_{7}',
                           'Y_{0}Y_{1}Z_{3}Z_{5}X_{6}X_{7}', 'X_{0}Y_{1}Z_{2}Y_{3}X_{4}Y_{5}Z_{6}']

        # Generate string rtepresentations
        str_rep = [pauli.__str__() for pauli in paulis]

        # Test that string representations were correctly generated
        assert str_rep == correct_str_rep, f'String representations were not correctly generated'

    def test_pauli_repr(self):
        """
        Test the function that generates a string representation of the Pauli operator object.

        The test consists in computing the string representation of a set of examples.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['XZX', 'YYYY', 'YZZXX', 'YYZZXX', 'XYZYXYZ']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Correct string representations
        correct_str_rep = ['PauliOp(XZX,(0, 1, 2))', 'PauliOp(YYYY,(0, 1, 4, 5))', 'PauliOp(YZZXX,(0, 2, 4, 5, 7))',
                           'PauliOp(YYZZXX,(0, 1, 3, 5, 6, 7))', 'PauliOp(XYZYXYZ,(0, 1, 2, 3, 4, 5, 6))']

        # Generate string rtepresentations
        str_rep = [pauli.__repr__() for pauli in paulis]

        # Test that string representations were correctly generated
        assert str_rep == correct_str_rep, f'String representations for Pauli operator object were not correctly generated'

    def test_pauli_mul(self):
        """
        Test the functions that multiply two Pauli operators.

        The test consists in multipliying a set of example Pauli operators.
        """

        # Define input examples of indices and strings
        input_indices = [(0, 1, 2), (0, 1, 4, 5), (0, 2, 4, 5, 7),
                         (0, 1, 3, 5, 6, 7), (0, 1, 2, 3, 4, 5, 6)]
        input_strings = ['XZX', 'YYYY', 'YZZXX', 'YYZZXX', 'XYZYXYX']

        # Create Pauli operators from input examples
        paulis = [PauliOp(string, index)
                  for string, index in zip(input_strings, input_indices)]

        # Compute multiplication between contiguous Pauli operators in the list
        pauli_mult = [paulis[j].__mul__(paulis[j+1])
                      for j in range(len(paulis)-1)]
        indices = [pauli.qubit_indices for pauli in pauli_mult]
        strings = [pauli.pauli_str for pauli in pauli_mult]
        phases = [pauli.phase for pauli in pauli_mult]

        # Correct matrix multiplications
        correct_indices = [(0, 1, 2, 4, 5), (1, 2, 4, 5, 7),
                           (1, 2, 3, 4, 5, 6), (0, 2, 3, 4, 5, 7)]
        correct_strings = ['ZXXYY', 'YZXZX', 'YZZZYX', 'ZZXXXX']
        correct_phases = [1, 1, -1j, 1j]

        # Test that matrix multiplications are correct
        assert indices == correct_indices, f'Indices from matrix multiplications are not correct'
        assert strings == correct_strings, f'Strings from matrix multiplications are not correct'
        assert np.allclose(
            phases, correct_phases), f'Phases from matrix multiplications are not correct'

    def test_pauli_XYZI(self):
        """
        Test the functions that define each Pauli matrix.

        The test consists in checking the properties of each Pauli matrix when defined as a Pauli
        operator object.
        """

        # Define a qubit index to define the Pauli operator object
        index = 1

        # Define the Pauli matrices
        X = PauliOp.X(index)
        Y = PauliOp.Y(index)
        Z = PauliOp.Z(index)
        I = PauliOp.I(index)
        paulis = [X, Y, Z, I]

        # Extract attributes
        strings = [pauli.pauli_str for pauli in paulis]
        indices = [pauli.qubit_indices for pauli in paulis]
        phases = [pauli.phase for pauli in paulis]
        string_rep = [pauli.__str__() for pauli in paulis]
        string_obj_rep = [pauli.__repr__() for pauli in paulis]
        matrix_rep = [pauli.matrix for pauli in paulis]

        # Correct attributes
        correct_strings = ['X', 'Y', 'Z', 'I']
        correct_indices = [(1,), (1,), (1,), (1,)]
        correct_phases = [1, 1, 1, 1]
        correct_string_rep = ['X_{1}', 'Y_{1}', 'Z_{1}', 'I_{1}']
        correct_string_obj_rep = [
            'PauliOp(X,(1,))', 'PauliOp(Y,(1,))', 'PauliOp(Z,(1,))', 'PauliOp(I,(1,))']
        correct_matrix_rep = [np.array(([0, 1], [1, 0]), dtype=complex), np.array(([0, -1j], [1j, 0]), dtype=complex),
                              np.array(([1, 0], [0, -1]), dtype=complex), np.array(([1, 0], [0, 1]), dtype=complex)]

        # Check indices, strings, phases and matrix representations
        assert strings == correct_strings, f'Strings are not correctly defined for Pauli matrices'
        assert indices == correct_indices, f'Indices are not correctly defined for Pauli matrices'
        assert np.allclose(
            phases, correct_phases), f'Phases are not correctly defined for Pauli matrices'
        assert string_rep == correct_string_rep, f'String representations are not correctly defined for Pauli matrices'
        assert string_obj_rep == correct_string_obj_rep, f'String object representations are not correctly defined for Pauli matrices'
        assert np.all([(matrix_rep[i] == correct_matrix_rep[i]).all() for i in range(len(
            matrix_rep))]), f'Matrix representations are not correctly defined for Pauli matrices'

    def test_hamiltonian_init(self):
        """
        Test initialization method of the Hamiltonian class.
        
        The tests consists in checking correct initializations for representative examples.
        """

        # Number of qubits in the register
        n_qubits = 3

        ## First example

        # Define Pauli terms using shifted indices
        idx_ref = 3
        indices_ref = list(range(idx_ref,n_qubits+idx_ref))
        trivial_indices = [(i,) for i in indices_ref]
        singlet_indices = [(i,) for i in indices_ref]
        pair_indices = [(i, j) for j in indices_ref for i in range(idx_ref,j)]

        trivial_terms = [PauliOp('I', indices) for indices in trivial_indices]
        linear_terms = [PauliOp('Z', indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp('ZZ', indices) for indices in pair_indices]

        terms = trivial_terms + linear_terms + quadratic_terms

        # Define coefficients and constant
        trivial_coeffs = [0.5 for _ in range(len(trivial_indices))]
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = trivial_coeffs + linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Correct attributes
        correct_indices = list(range(n_qubits))
        correct_singlet_indices = [(i,) for i in correct_indices]
        correct_pair_indices = [(i, j) for j in correct_indices for i in range(j)]
        correct_linear_terms = [PauliOp('Z', indices) for indices in correct_singlet_indices]
        correct_quadratic_terms = [PauliOp('ZZ', indices) for indices in correct_pair_indices]
        correct_linear_coeffs = [-1 for _ in range(len(correct_singlet_indices))]
        correct_quadratic_coeffs = [0.5 for _ in range(len(correct_pair_indices))]

        correct_terms = correct_linear_terms + correct_quadratic_terms
        correct_coefficients = correct_linear_coeffs + correct_quadratic_coeffs
        correct_constant = constant + sum(trivial_coeffs)

        # Test Hamiltonian was defined correctly
        assert hamiltonian.terms  == correct_terms, f'Terms in the Hamiltonian were not constructed correctly'
        assert hamiltonian.coeffs == correct_coefficients, f'Coefficients in the Hamiltonian were not constructed correctly'
        assert hamiltonian.constant == correct_constant, f'Constant in the Hamiltonian was not correctly constructed'

        ## Second example

        # Define Pauli terms
        trivial_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]

        trivial_terms = [PauliOp('I', indices) for indices in trivial_indices]
        terms_strings = ['XX', 'YY']
        quadratic_terms = [PauliOp(string, indices)
                 for indices in pair_indices for string in terms_strings]

        terms = trivial_terms + quadratic_terms

        # Define coefficients and constant
        trivial_coeffs = [-1 for _ in range(len(trivial_terms))]
        quadratic_coeffs = [0.5 for _ in range(len(quadratic_terms))]
        coefficients = trivial_coeffs + quadratic_coeffs

        constant = 2
        
        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Correct attributes
        correct_terms = quadratic_terms
        correct_coefficients = quadratic_coeffs
        correct_constant = constant + sum(trivial_coeffs)

        # Test Hamiltonian was defined correctly
        assert hamiltonian.terms  == correct_terms, f'Terms in the Hamiltonian were not constructed correctly'
        assert hamiltonian.coeffs == correct_coefficients, f'Coefficients in the Hamiltonian were not constructed correctly'
        assert hamiltonian.constant == correct_constant, f'Constant in the Hamiltonian was not correctly constructed'

        ## Exception case - number of terms does not match number of coefficients
        
        # Number of qubits
        n_qubits = 5

        # Define terms and coefficients
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms = [PauliOp('ZZ', indices) for indices in pair_indices]
        coefficients = [1 for _ in range(len(terms)+1)]

        # Attempt construction of Hamiltonian
        with self.assertRaises(AssertionError) as context:
            hamiltonian = Hamiltonian(terms, coefficients, constant = 0)
            
        # Check exception message
        self.assertEqual("Number of Pauli terms in Hamiltonian should be same as number of coefficients", str(context.exception))

        ## Exception case - non Pauli object term present in list of terms
        
        # Number of qubits
        n_qubits = 5

        # Define terms and coefficients
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        error_term = (0,5)
        terms = [PauliOp('ZZ', indices) for indices in pair_indices] + [error_term]
        coefficients = [1 for _ in range(len(terms))]
        
        # Attempt construction of Hamiltonian
        with self.assertRaises(TypeError) as context:
            hamiltonian = Hamiltonian(terms, coefficients, constant = 0)

        # Check exception message
        self.assertEqual(f"Pauli terms should be of type PauliOp and not {type(error_term)}", str(context.exception))
        
    def test_hamiltonian_qureg(self):
        """
        Tests the function that generated the register in the Hamiltonian object.

        The test consists in generating registers from a set of examples.
        """

        # Number of qubits in the register
        n_qubits = 5

        # Define Pauli terms
        terms_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms_strings = ['XX', 'YY']
        terms = [PauliOp(string, indices)
                 for indices in terms_indices for string in terms_strings]

        # Define coefficients and constant
        coefficients = [1 for _ in range(len(terms))]
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract physical register
        register = hamiltonian.qureg

        # Correct register
        correct_register = [0, 1, 2, 3, 4]

        # Test if the extracted register is correct
        assert np.allclose(
            register, correct_register), f'Register has been incorrectly constructed'

    def test_hamiltonian_divide_into_singlets_pairs(self):
        """
        Tests the function that divides that Hamiltonian into singlets and pairs.

        The tests consists in splittig terms into singlets and pairs from a given Hamiltonian.
        """

        # Number of qubits in the register
        n_qubits = 5

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        correct_linear_terms = [PauliOp('Z', indices)
                                for indices in singlet_indices]
        correct_quadratic_terms = [
            PauliOp('ZZ', indices) for indices in pair_indices]

        constant_term = [PauliOp('I', (0,))]

        terms = correct_linear_terms + correct_quadratic_terms + constant_term

        # Define coefficients and constant
        correct_linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        correct_quadratic_coeffs = [1 for _ in range(len(pair_indices))]
        constant_coeff = [1]
        coefficients = correct_linear_coeffs + correct_quadratic_coeffs + constant_coeff

        constant = 1
        correct_constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract singlets and pairs attributes
        linear_terms = hamiltonian.qubits_singles
        quadratic_terms = hamiltonian.qubits_pairs
        linear_coeffs = hamiltonian.single_qubit_coeffs
        quadratic_coeffs = hamiltonian.pair_qubit_coeffs
        constant = hamiltonian.constant

        # Test if the hamiltonian terms and coefficients have been correctly separated
        assert linear_terms == correct_linear_terms, f'Linear terms have not been correctly generated'
        assert np.allclose(
            linear_coeffs, correct_linear_coeffs), f'Linear coefficients have not been correctly generated'
        assert quadratic_terms == correct_quadratic_terms, f'Quadratic terms have not been correctly generated'
        assert np.allclose(
            quadratic_coeffs, correct_quadratic_coeffs), f'Quadratic coefficients have not been correctly generated'
        assert np.allclose(
            constant, correct_constant), f'Constant coefficient have not been correctly generated'

        # Exception case - generating terms beyond quadratic order

        # Define hamiltonian attributes
        terms_exc = [PauliOp('Z',(0,)), PauliOp('ZZ',(0,1)), PauliOp('ZZ',(0,2)),
                    PauliOp('ZZ',(1,2)), PauliOp('ZZZ',(1,2,3))]
        coefficients_exc = [1,0.5,0.5,0.5,3]
        constant_exc = 2

        # Test the exception is raised - Attempt construction of Hamiltonian
        with self.assertRaises(NotImplementedError) as context:
            hamiltonian_exc = Hamiltonian(terms_exc,coefficients_exc,constant_exc)

        # Check exception message
        self.assertEqual("Hamiltonian only supports Linear and Quadratic terms", str(context.exception))
        
    def test_hamiltonian_str(self):
        """
        Tests the function that constructs a string representation of the Hamiltonian object.

        The test consists of constructing the string representation of a given Hamiltonian.
        """

        # Number of qubits in the register
        n_qubits = 3

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp('Z', indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp('ZZ', indices) for indices in pair_indices]

        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract string representation
        string_rep = hamiltonian.__str__()

        # Correct string representation
        correct_string_rep = '-1*Z_{0} + -1*Z_{1} + -1*Z_{2} + 0.5*Z_{0}Z_{1} + 0.5*Z_{0}Z_{2} + 0.5*Z_{1}Z_{2} + 2'

        # Test the obtained string representation
        assert string_rep == correct_string_rep, f'Hamiltonian string representation was incorrectly constructed'

    def test_hamiltonian_len(self):
        """
        Tests the function that computes the number of terms in the Hamiltonian.

        The test consists in checking the number of terms for a fully connected Hamiltonian
        with linear terms for an increasing number of qubits
        """

        # Initialize lists containing correct and computed lengths
        correct_lengths = []
        lengths = []

        for n_qubits in range(3, 10):

            # Compute correct lengths
            correct_lengths.append(n_qubits + n_qubits*(n_qubits-1)/2)

            # Define Pauli terms
            singlet_indices = [(i,) for i in range(n_qubits)]
            pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
            linear_terms = [PauliOp('Z', indices)
                            for indices in singlet_indices]
            quadratic_terms = [PauliOp('ZZ', indices)
                               for indices in pair_indices]

            terms = linear_terms + quadratic_terms

            # Define coefficients and constant
            linear_coeffs = [-1 for _ in range(len(singlet_indices))]
            quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
            coefficients = linear_coeffs + quadratic_coeffs

            constant = 2

            # Define Hamiltonian
            hamiltonian = Hamiltonian(terms, coefficients, constant)

            # Compute length
            length = hamiltonian.__len__()
            lengths.append(length)

        # Test that computed lengths are correct
        assert np.allclose(
            lengths, correct_lengths), f'Hamiltonian lengths are incorrectly computed'

    def test_hamiltonian_expression(self):
        """
        Tests the function that computes the symbolic expression of the Hamiltonian object.

        The test consists in computing the symbolic expression for a specific example.
        """

        # Number of qubits in the register
        n_qubits = 3

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp('Z', indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp('ZZ', indices) for indices in pair_indices]

        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs

        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Extract symbolic representation
        sym_rep = hamiltonian.expression

        # Correct symbolic representation
        correct_sym_rep = Symbol(str(constant))

        for ind, coeff in enumerate(linear_coeffs):
            correct_sym_rep += Symbol(str(coeff)+'Z'+'_{'+str(ind)+'}')

        for ind, coeff in enumerate(quadratic_coeffs):
            i, j = pair_indices[ind]
            correct_sym_rep += Symbol(str(coeff) +
                                      'Z'+'_{'+str(i)+'}'+'Z'+'_{'+str(j)+'}')

        # Test if string representation was correctly generated
        assert sym_rep == correct_sym_rep, f'Symbolic representation was incorrectly generated'

    def test_hamiltonian_add(self):
        """
        Test the function that adds two Hamiltonian objects.

        The test consists in adding two given Hamiltonians with a repeated term.
        """

        # Number of qubits in the register
        n_qubits = 4

        # Define the first Hamiltonian
        pair_indices = [(i, j) for j in range(n_qubits-1) for i in range(j)]
        input_terms = [PauliOp('ZZ', indices) for indices in pair_indices]

        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        input_coefficients = quadratic_coeffs
        input_constant = 2

        hamiltonian = Hamiltonian(
            input_terms, input_coefficients, input_constant)

        # Define the other Hamiltonian
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp('Z', indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp('ZZ', indices) for indices in pair_indices]
        other_input_terms = linear_terms + quadratic_terms

        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(
            int((n_qubits-1)*(n_qubits-2)/2))] + [1 for _ in range(n_qubits-1)]
        other_input_coefficients = linear_coeffs + quadratic_coeffs

        other_input_constant = 0

        other_hamiltonian = Hamiltonian(
            other_input_terms, other_input_coefficients, other_input_constant)

        # Add hamiltonians
        hamiltonian.__add__(other_hamiltonian)

        # Exctract terms, coefficients and constants from added Hamiltonians
        terms = hamiltonian.terms
        coefficients = hamiltonian.coeffs
        constant = hamiltonian.constant

        # Create Hamiltonian dictionary - To be removed upon implementation of _simplify()
        hamiltonian_dict = {term.qubit_indices: coeff for term,
                            coeff in zip(hamiltonian.terms, hamiltonian.coeffs)}

        # Correct terms, coefficients and constants
        correct_terms = input_terms + other_input_terms
        correct_coefficients = input_coefficients + other_input_coefficients
        correct_constant = input_constant + other_input_constant

        # Create Hamiltonian dictionary for correct Hamiltonian - To be removed upon implementation of _simplify()
        correct_hamiltonian_dict = {term.qubit_indices: coeff for term, coeff in zip(
            input_terms, input_coefficients)}

        # Add terms and coefficients from the other hamiltonian
        for term, coeff in zip(other_input_terms, other_input_coefficients):

            # If term was absent add it together with coefficient
            if correct_hamiltonian_dict.get(term.qubit_indices) is None:
                correct_hamiltonian_dict.update({term.qubit_indices: coeff})

            # If term was present add coefficient to current term
            else:
                correct_hamiltonian_dict[term.qubit_indices] += coeff

        # Test if Hamiltonians were added correctly
        assert hamiltonian_dict == correct_hamiltonian_dict, f'Hamiltonian adder did not yield correct terms and/or coefficients'
        assert np.allclose(constant, correct_constant), f'Hamiltonian adder did not yield correct constant'

    def test_hamiltonian_squared(self):
        """
        Tests the function that computes the squared Hamiltonian, also a Hamiltonian object.

        The test consists in computing the Hamiltonian squared for two given small examples.
        """

        # First example

        # Number of qubits in the register
        n_qubits = 2

        # Define Pauli terms
        singlet_indices = [(i,) for i in range(n_qubits)]
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        linear_terms = [PauliOp('Z', indices) for indices in singlet_indices]
        quadratic_terms = [PauliOp('ZZ', indices) for indices in pair_indices]
        terms = linear_terms + quadratic_terms

        # Define coefficients and constant
        linear_coeffs = [-1 for _ in range(len(singlet_indices))]
        quadratic_coeffs = [0.5 for _ in range(len(pair_indices))]
        coefficients = linear_coeffs + quadratic_coeffs
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Compute squared hamiltonian and extract attributes
        hamiltonian_sq = hamiltonian.hamiltonian_squared

        terms_sq = hamiltonian_sq.terms
        coefficients_sq = hamiltonian_sq.coeffs
        constant_sq = hamiltonian_sq.constant

        # Correct Hamiltonian squared attributes
        correct_terms_sq = [PauliOp('ZZ', (0, 1)), PauliOp('Z', (1,)),
                            PauliOp('ZZ', (0, 1)), PauliOp('Z', (0,)),
                            PauliOp('Z', (1,)),  PauliOp('Z', (0,)),
                            PauliOp('Z', (0,)), PauliOp('Z', (1,)),
                            PauliOp('ZZ', (0, 1))]

        correct_coefficients_sq = [1, -0.5, 1, -0.5, -0.5, -0.5, -4, -4, 2]

        correct_constant_sq = 4 + n_qubits * \
            (-1)**2 + n_qubits*(n_qubits-1)/2*(0.5)**2

        # Test if Hamiltonian squared was correctly computed
        assert terms_sq == correct_terms_sq, f'Hamiltonian squared did not yield correct terms'
        assert np.allclose(
            coefficients_sq, correct_coefficients_sq), f'Hamiltonian squared did not yield correct coefficients'
        assert np.allclose(
            constant_sq, correct_constant_sq), f'Hamiltonian squared did not yield correct constant'

        # Second example

        # Number of qubits in the register
        n_qubits = 2

        # Define Pauli terms
        pair_indices = [(i, j) for j in range(n_qubits) for i in range(j)]
        terms_strings = ['XX', 'YY']
        terms = [PauliOp(string, indices)
                 for indices in pair_indices for string in terms_strings]

        # Define coefficients and constant
        coefficients = [1 for _ in range(len(terms))]
        constant = 2

        # Define Hamiltonian
        hamiltonian = Hamiltonian(terms, coefficients, constant)

        # Compute squared hamiltonian and extract attributes
        hamiltonian_sq = hamiltonian.hamiltonian_squared

        terms_sq = hamiltonian_sq.terms
        coefficients_sq = hamiltonian_sq.coeffs
        constant_sq = hamiltonian_sq.constant

        # Correct Hamiltonian squared attributes
        correct_terms_sq = [PauliOp('ZZ', (0, 1)), PauliOp(
            'ZZ', (0, 1)), PauliOp('XX', (0, 1)), PauliOp('YY', (0, 1))]
        correct_coefficients_sq = [-1, -1, 4, 4]
        correct_constant_sq = 6

        # Test if Hamiltonian squared was correctly computed
        assert terms_sq == correct_terms_sq, f'Hamiltonian squared did not yield correct terms'
        assert np.allclose(
            coefficients_sq, correct_coefficients_sq), f'Hamiltonian squared did not yield correct coefficients'
        assert np.allclose(
            constant_sq, correct_constant_sq), f'Hamiltonian squared did not yield correct constant'

    def test_classical_hamiltonian(self):
        """
        Tests the function that generates a classical Hamiltonian, i.e. a Hamiltonian composed only
        of Z Pauli operators.

        The tests consists in generating a Hamiltonian from a given set of graph edges and weights.
        """

        # Number of nodes in the graph, i.e. qubits in the register
        n_nodes = 5

        # Graph properties
        input_edges = [(i,) for i in range(n_nodes)] + [(i, j)
                                                        for j in range(n_nodes) for i in range(j)]
        input_weights = [-1 for _ in range(len(input_edges))]
        input_constant = 2

        # Generate classical Hamiltonian and extract attributes
        cl_hamiltonian = Hamiltonian.classical_hamiltonian(
            input_edges, input_weights, input_constant)

        terms = cl_hamiltonian.terms
        terms_indices = [term.qubit_indices for term in terms]
        coefficients = cl_hamiltonian.coeffs
        constant = cl_hamiltonian.constant

        # Test that the generated Hamiltonian match the properties of the input graph
        assert terms_indices == input_edges, f'Classical Hamiltonian terms did not match input graph edges'
        assert np.allclose(
            coefficients, input_weights), f'Hamiltonian squared did not yield correct coefficients'
        assert np.allclose(
            constant, input_constant), f'Hamiltonian squared did not yield correct constant'

        # Exception case - inserting coefficients that are not interegers or floats

        # Define hamiltonian attributes
        terms_exc = [(0,1),(0,2),(1,2)]
        constant_exc = 2

        # Define different types of coefficients containing exceptions
        coefficients_exc_types = [[-1j,1,-1j],[2,3,0.1*1j]]
        
        # Check for every coefficient type
        for coefficients_exc in coefficients_exc_types:
            
            # Test the exception is raised - Attempt construction of Hamiltonian
            with self.assertRaises(ValueError) as context:
                hamiltonian_exc = Hamiltonian.classical_hamiltonian(terms_exc,coefficients_exc,constant_exc)
            
        # Check exception message
        self.assertEqual("Classical Hamiltonians only support Integer or Float coefficients", str(context.exception))
        
        # Exception case - generating terms beyond quadratic order

        # Define hamiltonian attributes
        terms_exc = [(0,),(0,1),(0,2),(1,2),(1,2,3)]
        coefficients_exc = [1,0.5,0.5,0.5,3]
        constant_exc = 2

        # Test the exception is raised - Attempt construction of Hamiltonian
        with self.assertRaises(ValueError) as context:
            hamiltonian_exc = Hamiltonian.classical_hamiltonian(terms_exc,coefficients_exc,constant_exc)

        # Check exception message
        self.assertEqual("Hamiltonian only supports Linear and Quadratic terms", str(context.exception))

if __name__ == "__main__":
    unittest.main()


from setuptools import setup, find_namespace_packages
from os import getcwd

current_path = getcwd()

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("src/openqaoa-core/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = [
    "amazon-braket-sdk>=1.23.0",
    "pandas>=1.3.5",
    "sympy>=1.10.1",
    "numpy>=1.22.3",
    "networkx>=2.8",
    "matplotlib>=3.4.3",
    "scipy>=1.8",
    "qiskit>=0.36.1",
    "pyquil>=3.1.0",
    "docplex>=2.23.1",
    "autograd>=1.4",
    "semantic_version>=2.10",
    "autoray>=0.3.1",
    "azure-quantum",
    "qdk",
    "qiskit-qir",
    "qiskit-ionq",
    "azure-quantum[qiskit]",
]

requirements_docs = [
    "sphinx>=4.5.0",
    "sphinx-autodoc-typehints>=1.18.1",
    "sphinx-rtd-theme>=1.0.0",
    "nbsphinx>=0.8.9",
    "ipython>=8.10.0",
    "nbconvert>=6.5.1",
]

requirements_test = [
    "pytest>=7.1.0",
    "pytest-cov>=3.0.0",
    "ipython>=8.2.0",
    "nbconvert>=6.5.1",
    "pandas>=1.4.3",
    "plotly>=5.9.0",
    "cplex>=22.1.0.0",
]

package_names = [
    "openqaoa",
    "openqaoa_braket",
    "openqaoa_qiskit",
    "openqaoa_pyquil",
    "openqaoa_azure",
]
folder_names = [
    "openqaoa-core",
    "openqaoa-braket",
    "openqaoa-qiskit",
    "openqaoa-pyquil",
    "openqaoa-azure",
]
packages_import = find_namespace_packages(where="./src")
updated_packages = []
for each_package_name in packages_import:
    for _index, each_folder_name in enumerate(folder_names):
        if each_folder_name in each_package_name:
            updated_packages.append(
                each_package_name.replace(each_folder_name, package_names[_index])
            )
            continue

setup(
    name="openqaoa",
    python_requires=">=3.8, <3.11",
    version=version,
    author="Entropica Labs",
    packages=updated_packages,
    package_dir={
        "": "src",
        "openqaoa": "src/openqaoa-core",
        "openqaoa_braket": "src/openqaoa-braket",
        "openqaoa_qiskit": "src/openqaoa-qiskit",
        "openqaoa_pyquil": "src/openqaoa-pyquil",
        "openqaoa_azure": "src/openqaoa-azure",
    },
    url="https://github.com/entropicalabs/openqaoa",
    install_requires=requirements,
    license="MIT",
    description="OpenQAOA is a python open-source multi-backend Software Development Kit to create, customise and execute the Quantum Approximate Optimisation Algorithm (QAOA) on Noisy Intermediate-Scale Quantum (NISQ) devices, and simulators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    keywords="quantum optimisation SDK",
    extras_require={
        "docs": requirements_docs,
        "tests": requirements_test,
        "all": requirements_docs + requirements_test,
    },
)
